#!/usr/local/bin/perl -wT
# Copyright Andru Luvisi, 1999
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


use CGI;
use CGI::Carp qw(fatalsToBrowser);
use English;
use strict;
use vars qw($query @words @wordmasks $fodder $fodmask $fodlen @anagrams);
use vars qw($wordfile $header $footer $minwordlen $maxwords $tmplen %ENV);
use vars qw($zeromask $tmpmask $maskfile $word $numwords $offset $mask);
use vars qw($| $i $j);

$wordfile = "/usr/dict/words";
$maskfile = "anamasks.dat";
$ENV{"PATH"} = "/bin:/usr/bin";
$|=1; # so the user gets *something* back soon...

# my /usr/dict/words has 45402 words...  adjust as needed.
$#words = $#wordmasks = 50000; 

$zeromask = "\000" x 52;

$header = <<EOT;
 <title>Anagram Results</title>
 <h1>Anagram Results</h1>
EOT

sub print_anagram {
 print join(" ", @_), "<br>\n";
}

# We would be grateful if you could leave this link in, but
# we understand if you can't.  :-)
#    - Villainy Amusements
$footer = <<EOT;
 <font size="-1">
  <a href="http://www.villainy.com/">
   Anagram Amusement created by Villainy Amusements.
  </a>
 </font>
EOT


$query = CGI->new();

$minwordlen = $query->param("minwordlen") || 4;
$maxwords = $query->param("maxwords") || 4;

$fodder = $query->param("fodder");
$fodder =~ tr/A-Z/a-z/;
$fodder =~ s/[^a-z]//g;
$fodlen = length($fodder);
$fodmask = makemask($fodder, $minwordlen, $fodlen);

if($fodlen < $minwordlen) {
 print $query->header(), 
  "<p>You need to enter at least $minwordlen letters if you are only " .
  "going to accept anagrams containing at least $minwordlen letters.\n";
 exit(0);
}

#if(!open(WORDS, "-|")) {
# exec("grep", "-e", "^[$fodder]\\{$minwordlen,$fodlen\\}\$", $wordfile) ||
#  (print("Couldn't exec grep"), die("Couldn't exec grep"));
#}
open(WORDS, $wordfile) || die("Can't open $wordfile");
open(MASKS, $maskfile) || die("Can't open $maskfile");

seek(MASKS, 0, 2);
$numwords = tell(MASKS)/64;

for($i = 0, $j = -1; $i < $numwords; $i++) {
 seek(MASKS, $i*64, 0);
 read(MASKS, $tmpmask, 60);
 if(($tmpmask & $fodmask) eq $tmpmask) {
  read(MASKS, $offset, 4);
  $offset = unpack("L", $offset);
  seek(WORDS, $offset, 0);
  $word = <WORDS>;
  chop($word);
  $words[++$j] = $word;
  $wordmasks[$j] = $tmpmask;
 }
}

close(MASKS);
close(WORDS);
$#words = $#wordmasks = $j;
fyshuffle(\@words, \@wordmasks);

@anagrams = ();
print $query->header(), $header;
find_anagrams($fodder, 0, \&print_anagram);
print $footer, "\n";
exit(0);


# find_word($superword, $startindex)
sub find_word {
 my($superword, $startindex) = @_;
 my($i, $mask);

 if(length($superword) < $minwordlen) {
  return -1;
 }

 $mask = makemask($superword, 0, length($superword));
 for($i = $startindex; $i < @words; $i++) {
  if(($wordmasks[$i] & $mask) eq $wordmasks[$i]) {
   return $i;
  }
 }
 return -1;
}

# find_anagrams($word, $startindex, $callback, @found);
sub find_anagrams {
 my ($word, $startindex) = (shift, shift);
 my ($callback) = shift;
 my @found = @_;
 my $index;

 if(length($word) < $minwordlen || @found >= $maxwords || 
    ($index = find_word($word, $startindex)) == -1) {
  &$callback(@found);
  return;
 }

 while($index != -1) {
  my $tempword = $word;
  my $foundword = $words[$index];

  while($foundword =~ /[A-Za-z]/g) {
   my $remove = $&;
   $tempword =~ s/$remove//;
  }
  find_anagrams($tempword, $index + 1, $callback, @found, $foundword);
  $index = find_word($word, $index + 1);
 }
 return;
}

# create a bitmask of the letters used in the word
sub makemask {
 my $word = shift;
 my $minlen = shift;
 my $maxlen = shift;
 my $mask = "\000" x 60;
 my $offset;

 while($word =~ /[A-Za-z]/g) {
  $offset = (ord(lc($&)) - 97);
  vec($mask, $offset, 16) = 1 | (vec($mask, $offset, 16) << 1);
 }
 vec($mask, 13, 32) = ~0 << $minlen;      # encode length for minimum testing
 vec($mask, 14, 32) = ~0 >> 32 - $maxlen; # encode length for maximum testing

 return $mask;
}



# is_contained(word1, word2);  is word1 contained in word2?  brute force.
sub is_contained {
 my ($word1, $word2) = @_;

 foreach my $letter (split(//, $word1)) {
  $word2 =~ s/$letter// || return 0;
 }

 return 1;
}

sub fyshuffle {
 my($array1, $array2) = @_;
 my $i;

 return if !@$array1;
 for($i = @$array1; --$i;) {
  my($j) = int(rand($i+1));
  next if $i == $j;
  @$array1[$i,$j] = @$array1[$j,$i];
  @$array2[$i,$j] = @$array2[$j,$i];
 }
}

